@@ -387,8 +387,8 @@ int Client::quic_init(const sockaddr *local_addr, socklen_t local_addrlen,
     settings.qlog.write = qlog_write_cb;
   }
   if (config->max_udp_payload_size) {
-    settings.max_udp_payload_size = config->max_udp_payload_size;
-    settings.no_udp_payload_size_shaping = 1;
+    settings.max_tx_udp_payload_size = config->max_udp_payload_size;
+    settings.no_tx_udp_payload_size_shaping = 1;
   }
 
   ngtcp2_transport_params params;
@@ -496,8 +496,7 @@ int Client::quic_pkt_timeout() {
 
   rv = ngtcp2_conn_handle_expiry(quic.conn, now);
   if (rv != 0) {
-    ngtcp2_connection_close_error_set_transport_error_liberr(&quic.last_error,
-                                                             rv, nullptr, 0);
+    ngtcp2_ccerr_set_liberr(&quic.last_error, rv, nullptr, 0);
     return -1;
   }
 
@@ -550,12 +549,11 @@ int Client::read_quic() {
 
       if (!quic.last_error.error_code) {
         if (rv == NGTCP2_ERR_CRYPTO) {
-          ngtcp2_connection_close_error_set_transport_error_tls_alert(
-              &quic.last_error, ngtcp2_conn_get_tls_alert(quic.conn), nullptr,
-              0);
+          ngtcp2_ccerr_set_tls_alert(&quic.last_error,
+                                     ngtcp2_conn_get_tls_alert(quic.conn),
+                                     nullptr, 0);
         } else {
-          ngtcp2_connection_close_error_set_transport_error_liberr(
-              &quic.last_error, rv, nullptr, 0);
+          ngtcp2_ccerr_set_liberr(&quic.last_error, rv, nullptr, 0);
         }
       }
 
@@ -592,14 +590,14 @@ int Client::write_quic() {
 
   std::array<nghttp3_vec, 16> vec;
   size_t pktcnt = 0;
-  auto max_udp_payload_size = ngtcp2_conn_get_max_udp_payload_size(quic.conn);
+  auto max_udp_payload_size =
+      ngtcp2_conn_get_max_tx_udp_payload_size(quic.conn);
 #ifdef UDP_SEGMENT
   auto path_max_udp_payload_size =
-      ngtcp2_conn_get_path_max_udp_payload_size(quic.conn);
+      ngtcp2_conn_get_path_max_tx_udp_payload_size(quic.conn);
 #endif // UDP_SEGMENT
-  size_t max_pktcnt =
-      std::min(static_cast<size_t>(10),
-               static_cast<size_t>(64_k / max_udp_payload_size));
+  auto max_pktcnt =
+      ngtcp2_conn_get_send_quantum(quic.conn) / max_udp_payload_size;
   uint8_t *bufpos = quic.tx.data.get();
   ngtcp2_path_storage ps;
   size_t gso_size = 0;
@@ -637,15 +635,11 @@ int Client::write_quic() {
       switch (nwrite) {
       case NGTCP2_ERR_STREAM_DATA_BLOCKED:
         assert(ndatalen == -1);
-        if (s->block_stream(stream_id) != 0) {
-          return -1;
-        }
+        s->block_stream(stream_id);
         continue;
       case NGTCP2_ERR_STREAM_SHUT_WR:
         assert(ndatalen == -1);
-        if (s->shutdown_stream_write(stream_id) != 0) {
-          return -1;
-        }
+        s->shutdown_stream_write(stream_id);
         continue;
       case NGTCP2_ERR_WRITE_MORE:
         assert(ndatalen >= 0);
@@ -655,8 +649,7 @@ int Client::write_quic() {
         continue;
       }
 
-      ngtcp2_connection_close_error_set_transport_error_liberr(
-          &quic.last_error, nwrite, nullptr, 0);
+      ngtcp2_ccerr_set_liberr(&quic.last_error, nwrite, nullptr, 0);
       return -1;
     } else if (ndatalen >= 0 && s->add_write_offset(stream_id, ndatalen) != 0) {
       return -1;
