@@ -49,22 +49,6 @@
 
 namespace shrpx {
 
-namespace {
-void idle_timeoutcb(struct ev_loop *loop, ev_timer *w, int revents) {
-  auto upstream = static_cast<Http3Upstream *>(w->data);
-
-  if (LOG_ENABLED(INFO)) {
-    ULOG(INFO, upstream) << "QUIC idle timeout";
-  }
-
-  upstream->idle_close();
-
-  auto handler = upstream->get_client_handler();
-
-  delete handler;
-}
-} // namespace
-
 namespace {
 void timeoutcb(struct ev_loop *loop, ev_timer *w, int revents) {
   auto upstream = static_cast<Http3Upstream *>(w->data);
@@ -116,30 +100,34 @@ size_t downstream_queue_size(Worker *worker) {
 }
 } // namespace
 
+namespace {
+ngtcp2_conn *get_conn(ngtcp2_crypto_conn_ref *conn_ref) {
+  auto conn = static_cast<Connection *>(conn_ref->user_data);
+  auto handler = static_cast<ClientHandler *>(conn->data);
+  auto upstream = static_cast<Http3Upstream *>(handler->get_upstream());
+  return upstream->get_conn();
+}
+} // namespace
+
 Http3Upstream::Http3Upstream(ClientHandler *handler)
     : handler_{handler},
-      max_udp_payload_size_{SHRPX_QUIC_MAX_UDP_PAYLOAD_SIZE},
       qlog_fd_{-1},
       hashed_scid_{},
       conn_{nullptr},
-      tls_alert_{0},
       httpconn_{nullptr},
       downstream_queue_{downstream_queue_size(handler->get_worker()),
                         !get_config()->http2_proxy},
-      idle_close_{false},
       retry_close_{false},
       tx_{
           .data = std::unique_ptr<uint8_t[]>(new uint8_t[64_k]),
       } {
+  auto conn = handler_->get_connection();
+  conn->conn_ref.get_conn = shrpx::get_conn;
+
   ev_timer_init(&timer_, timeoutcb, 0., 0.);
   timer_.data = this;
 
-  auto config = get_config();
-  auto &quicconf = config->quic;
-
-  ev_timer_init(&idle_timer_, idle_timeoutcb, 0.,
-                quicconf.upstream.timeout.idle);
-  idle_timer_.data = this;
+  ngtcp2_ccerr_default(&last_error_);
 
   ev_timer_init(&shutdown_timer_, shutdown_timeout_cb, 0., 0.);
   shutdown_timer_.data = this;
@@ -154,7 +142,6 @@ Http3Upstream::~Http3Upstream() {
 
   ev_prepare_stop(loop, &prep_);
   ev_timer_stop(loop, &shutdown_timer_);
-  ev_timer_stop(loop, &idle_timer_);
   ev_timer_stop(loop, &timer_);
 
   nghttp3_conn_del(httpconn_);
@@ -300,7 +287,9 @@ int Http3Upstream::recv_stream_data(uint32_t flags, int64_t stream_id,
   if (nconsumed < 0) {
     ULOG(ERROR, this) << "nghttp3_conn_read_stream: "
                       << nghttp3_strerror(nconsumed);
-    last_error_ = quic::err_application(nconsumed);
+    ngtcp2_ccerr_set_application_error(
+        &last_error_, nghttp3_err_infer_quic_app_error_code(nconsumed), nullptr,
+        0);
     return -1;
   }
 
@@ -344,7 +333,8 @@ int Http3Upstream::stream_close(int64_t stream_id, uint64_t app_error_code) {
     break;
   default:
     ULOG(ERROR, this) << "nghttp3_conn_close_stream: " << nghttp3_strerror(rv);
-    last_error_ = quic::err_application(rv);
+    ngtcp2_ccerr_set_application_error(
+        &last_error_, nghttp3_err_infer_quic_app_error_code(rv), nullptr, 0);
     return -1;
   }
 
@@ -489,21 +479,48 @@ int Http3Upstream::handshake_completed() {
     return -1;
   }
 
-  std::array<uint8_t, NGTCP2_CRYPTO_MAX_REGULAR_TOKENLEN> token;
+  auto path = ngtcp2_conn_get_path(conn_);
+
+  return send_new_token(&path->remote);
+}
+
+namespace {
+int path_validation(ngtcp2_conn *conn, uint32_t flags, const ngtcp2_path *path,
+                    const ngtcp2_path *old_path,
+                    ngtcp2_path_validation_result res, void *user_data) {
+  if (res != NGTCP2_PATH_VALIDATION_RESULT_SUCCESS ||
+      !(flags & NGTCP2_PATH_VALIDATION_FLAG_NEW_TOKEN)) {
+    return 0;
+  }
+
+  auto upstream = static_cast<Http3Upstream *>(user_data);
+  if (upstream->send_new_token(&path->remote) != 0) {
+    return NGTCP2_ERR_CALLBACK_FAILURE;
+  }
+
+  return 0;
+}
+} // namespace
+
+int Http3Upstream::send_new_token(const ngtcp2_addr *remote_addr) {
+  std::array<uint8_t, NGTCP2_CRYPTO_MAX_REGULAR_TOKENLEN + 1> token;
   size_t tokenlen;
 
-  auto path = ngtcp2_conn_get_path(conn_);
   auto worker = handler_->get_worker();
   auto conn_handler = worker->get_connection_handler();
   auto &qkms = conn_handler->get_quic_keying_materials();
   auto &qkm = qkms->keying_materials.front();
 
-  if (generate_token(token.data(), tokenlen, path->remote.addr,
-                     path->remote.addrlen, qkm.secret.data(),
+  if (generate_token(token.data(), tokenlen, remote_addr->addr,
+                     remote_addr->addrlen, qkm.secret.data(),
                      qkm.secret.size()) != 0) {
-    return 0;
+    return -1;
   }
 
+  assert(tokenlen == NGTCP2_CRYPTO_MAX_REGULAR_TOKENLEN);
+
+  token[tokenlen++] = qkm.id;
+
   auto rv = ngtcp2_conn_submit_new_token(conn_, token.data(), tokenlen);
   if (rv != 0) {
     ULOG(ERROR, this) << "ngtcp2_conn_submit_new_token: "
@@ -514,6 +531,21 @@ int Http3Upstream::handshake_completed() {
   return 0;
 }
 
+namespace {
+int recv_tx_key(ngtcp2_conn *conn, ngtcp2_crypto_level level, void *user_data) {
+  if (level != NGTCP2_CRYPTO_LEVEL_APPLICATION) {
+    return 0;
+  }
+
+  auto upstream = static_cast<Http3Upstream *>(user_data);
+  if (upstream->setup_httpconn() != 0) {
+    return NGTCP2_ERR_CALLBACK_FAILURE;
+  }
+
+  return 0;
+}
+} // namespace
+
 int Http3Upstream::init(const UpstreamAddr *faddr, const Address &remote_addr,
                         const Address &local_addr,
                         const ngtcp2_pkt_hd &initial_hd,
@@ -545,7 +577,7 @@ int Http3Upstream::init(const UpstreamAddr *faddr, const Address &remote_addr,
       get_new_connection_id,
       remove_connection_id,
       ngtcp2_crypto_update_key_cb,
-      nullptr, // path_validation
+      shrpx::path_validation,
       nullptr, // select_preferred_addr
       shrpx::stream_reset,
       shrpx::extend_max_remote_streams_bidi,
@@ -561,6 +593,9 @@ int Http3Upstream::init(const UpstreamAddr *faddr, const Address &remote_addr,
       nullptr, // lost_datagram
       ngtcp2_crypto_get_path_challenge_data_cb,
       shrpx::stream_stop_sending,
+      nullptr, // version_negotiation
+      nullptr, // recv_rx_key
+      shrpx::recv_tx_key,
   };
 
   auto config = get_config();
@@ -599,10 +634,10 @@ int Http3Upstream::init(const UpstreamAddr *faddr, const Address &remote_addr,
   settings.cc_algo = quicconf.upstream.congestion_controller;
   settings.max_window = http3conf.upstream.max_connection_window_size;
   settings.max_stream_window = http3conf.upstream.max_window_size;
-  settings.max_udp_payload_size = SHRPX_QUIC_MAX_UDP_PAYLOAD_SIZE;
-  settings.assume_symmetric_path = 1;
+  settings.max_tx_udp_payload_size = SHRPX_QUIC_MAX_UDP_PAYLOAD_SIZE;
   settings.rand_ctx.native_handle = &worker->get_randgen();
-  settings.token = ngtcp2_vec{const_cast<uint8_t *>(token), tokenlen};
+  settings.token = token;
+  settings.tokenlen = tokenlen;
 
   ngtcp2_transport_params params;
   ngtcp2_transport_params_default(&params);
@@ -617,24 +652,28 @@ int Http3Upstream::init(const UpstreamAddr *faddr, const Address &remote_addr,
 
 #ifdef OPENSSL_IS_BORINGSSL
   if (quicconf.upstream.early_data) {
-    ngtcp2_transport_params early_data_params{
-        .initial_max_stream_data_bidi_local =
-            params.initial_max_stream_data_bidi_local,
-        .initial_max_stream_data_bidi_remote =
-            params.initial_max_stream_data_bidi_remote,
-        .initial_max_stream_data_uni = params.initial_max_stream_data_uni,
-        .initial_max_data = params.initial_max_data,
-        .initial_max_streams_bidi = params.initial_max_streams_bidi,
-        .initial_max_streams_uni = params.initial_max_streams_uni,
-    };
+    ngtcp2_transport_params early_data_params;
+
+    ngtcp2_transport_params_default(&early_data_params);
+
+    early_data_params.initial_max_stream_data_bidi_local =
+        params.initial_max_stream_data_bidi_local;
+    early_data_params.initial_max_stream_data_bidi_remote =
+        params.initial_max_stream_data_bidi_remote;
+    early_data_params.initial_max_stream_data_uni =
+        params.initial_max_stream_data_uni;
+    early_data_params.initial_max_data = params.initial_max_data;
+    early_data_params.initial_max_streams_bidi =
+        params.initial_max_streams_bidi;
+    early_data_params.initial_max_streams_uni = params.initial_max_streams_uni;
 
     // TODO include HTTP/3 SETTINGS
 
     std::array<uint8_t, 128> quic_early_data_ctx;
 
-    auto quic_early_data_ctxlen = ngtcp2_encode_transport_params(
+    auto quic_early_data_ctxlen = ngtcp2_transport_params_encode(
         quic_early_data_ctx.data(), quic_early_data_ctx.size(),
-        NGTCP2_TRANSPORT_PARAMS_TYPE_ENCRYPTED_EXTENSIONS, &early_data_params);
+        &early_data_params);
 
     assert(quic_early_data_ctxlen > 0);
     assert(static_cast<size_t>(quic_early_data_ctxlen) <=
@@ -657,6 +696,8 @@ int Http3Upstream::init(const UpstreamAddr *faddr, const Address &remote_addr,
     params.original_dcid = initial_hd.dcid;
   }
 
+  params.original_dcid_present = 1;
+
   rv = generate_quic_stateless_reset_token(
       params.stateless_reset_token, scid, qkm.secret.data(), qkm.secret.size());
   if (rv != 0) {
@@ -716,10 +757,16 @@ int Http3Upstream::on_write() {
     }
   }
 
+  handler_->get_connection()->wlimit.stopw();
+
   if (write_streams() != 0) {
     return -1;
   }
 
+  if (httpconn_ && nghttp3_conn_is_drained(httpconn_)) {
+    return -1;
+  }
+
   reset_timer();
 
   return 0;
@@ -727,32 +774,23 @@ int Http3Upstream::on_write() {
 
 int Http3Upstream::write_streams() {
   std::array<nghttp3_vec, 16> vec;
-  auto max_udp_payload_size = std::min(
-      max_udp_payload_size_, ngtcp2_conn_get_path_max_udp_payload_size(conn_));
-  size_t max_pktcnt =
-      std::min(static_cast<size_t>(64_k), ngtcp2_conn_get_send_quantum(conn_)) /
-      max_udp_payload_size;
+  auto max_udp_payload_size = ngtcp2_conn_get_max_tx_udp_payload_size(conn_);
+#ifdef UDP_SEGMENT
+  auto path_max_udp_payload_size =
+      ngtcp2_conn_get_path_max_tx_udp_payload_size(conn_);
+#endif // UDP_SEGMENT
+  auto max_pktcnt = ngtcp2_conn_get_send_quantum(conn_) / max_udp_payload_size;
   ngtcp2_pkt_info pi, prev_pi;
   uint8_t *bufpos = tx_.data.get();
   ngtcp2_path_storage ps, prev_ps;
   size_t pktcnt = 0;
   int rv;
+  size_t gso_size = 0;
   auto ts = quic_timestamp();
 
   ngtcp2_path_storage_zero(&ps);
   ngtcp2_path_storage_zero(&prev_ps);
 
-  auto config = get_config();
-  auto &quicconf = config->quic;
-
-  switch (quicconf.upstream.congestion_controller) {
-  case NGTCP2_CC_ALGO_BBR:
-  case NGTCP2_CC_ALGO_BBR2:
-    break;
-  default:
-    max_pktcnt = std::min(max_pktcnt, static_cast<size_t>(10));
-  }
-
   for (;;) {
     int64_t stream_id = -1;
     int fin = 0;
@@ -764,7 +802,9 @@ int Http3Upstream::write_streams() {
       if (sveccnt < 0) {
         ULOG(ERROR, this) << "nghttp3_conn_writev_stream: "
                           << nghttp3_strerror(sveccnt);
-        last_error_ = quic::err_application(sveccnt);
+        ngtcp2_ccerr_set_application_error(
+            &last_error_, nghttp3_err_infer_quic_app_error_code(sveccnt),
+            nullptr, 0);
         return handle_error();
       }
     }
@@ -785,23 +825,11 @@ int Http3Upstream::write_streams() {
       switch (nwrite) {
       case NGTCP2_ERR_STREAM_DATA_BLOCKED:
         assert(ndatalen == -1);
-        rv = nghttp3_conn_block_stream(httpconn_, stream_id);
-        if (rv != 0) {
-          ULOG(ERROR, this)
-              << "nghttp3_conn_block_stream: " << nghttp3_strerror(rv);
-          last_error_ = quic::err_application(rv);
-          return handle_error();
-        }
+        nghttp3_conn_block_stream(httpconn_, stream_id);
         continue;
       case NGTCP2_ERR_STREAM_SHUT_WR:
         assert(ndatalen == -1);
-        rv = nghttp3_conn_shutdown_stream_write(httpconn_, stream_id);
-        if (rv != 0) {
-          ULOG(ERROR, this)
-              << "nghttp3_conn_shutdown_stream_write: " << nghttp3_strerror(rv);
-          last_error_ = quic::err_application(rv);
-          return handle_error();
-        }
+        nghttp3_conn_shutdown_stream_write(httpconn_, stream_id);
         continue;
       case NGTCP2_ERR_WRITE_MORE:
         assert(ndatalen >= 0);
@@ -809,7 +837,9 @@ int Http3Upstream::write_streams() {
         if (rv != 0) {
           ULOG(ERROR, this)
               << "nghttp3_conn_add_write_offset: " << nghttp3_strerror(rv);
-          last_error_ = quic::err_application(rv);
+          ngtcp2_ccerr_set_application_error(
+              &last_error_, nghttp3_err_infer_quic_app_error_code(rv), nullptr,
+              0);
           return handle_error();
         }
         continue;
@@ -820,7 +850,7 @@ int Http3Upstream::write_streams() {
       ULOG(ERROR, this) << "ngtcp2_conn_writev_stream: "
                         << ngtcp2_strerror(nwrite);
 
-      last_error_ = quic::err_transport(nwrite);
+      ngtcp2_ccerr_set_liberr(&last_error_, nwrite, nullptr, 0);
 
       return handle_error();
     } else if (ndatalen >= 0) {
@@ -828,7 +858,9 @@ int Http3Upstream::write_streams() {
       if (rv != 0) {
         ULOG(ERROR, this) << "nghttp3_conn_add_write_offset: "
                           << nghttp3_strerror(rv);
-        last_error_ = quic::err_application(rv);
+        ngtcp2_ccerr_set_application_error(
+            &last_error_, nghttp3_err_infer_quic_app_error_code(rv), nullptr,
+            0);
         return handle_error();
       }
     }
@@ -842,26 +874,17 @@ int Http3Upstream::write_streams() {
         rv = send_packet(faddr, prev_ps.path.remote.addr,
                          prev_ps.path.remote.addrlen, prev_ps.path.local.addr,
                          prev_ps.path.local.addrlen, prev_pi, data, datalen,
-                         max_udp_payload_size);
+                         gso_size);
         if (rv == SHRPX_ERR_SEND_BLOCKED) {
           on_send_blocked(faddr, prev_ps.path.remote, prev_ps.path.local,
-                          prev_pi, data, datalen, max_udp_payload_size);
-
-          ngtcp2_conn_update_pkt_tx_time(conn_, ts);
-          reset_idle_timer();
+                          prev_pi, data, datalen, gso_size);
 
           signal_write_upstream_addr(faddr);
-
-          return 0;
         }
-
-        reset_idle_timer();
       }
 
       ngtcp2_conn_update_pkt_tx_time(conn_, ts);
 
-      handler_->get_connection()->wlimit.stopw();
-
       return 0;
     }
 
@@ -871,8 +894,12 @@ int Http3Upstream::write_streams() {
     if (pktcnt == 0) {
       ngtcp2_path_copy(&prev_ps.path, &ps.path);
       prev_pi = pi;
+      gso_size = nwrite;
     } else if (!ngtcp2_path_eq(&prev_ps.path, &ps.path) ||
-               prev_pi.ecn != pi.ecn) {
+               prev_pi.ecn != pi.ecn ||
+               static_cast<size_t>(nwrite) > gso_size ||
+               (gso_size > path_max_udp_payload_size &&
+                static_cast<size_t>(nwrite) != gso_size)) {
       auto faddr = static_cast<UpstreamAddr *>(prev_ps.path.user_data);
       auto data = tx_.data.get();
       auto datalen = bufpos - data - nwrite;
@@ -880,15 +907,15 @@ int Http3Upstream::write_streams() {
       rv = send_packet(faddr, prev_ps.path.remote.addr,
                        prev_ps.path.remote.addrlen, prev_ps.path.local.addr,
                        prev_ps.path.local.addrlen, prev_pi, data, datalen,
-                       max_udp_payload_size);
+                       gso_size);
       switch (rv) {
       case SHRPX_ERR_SEND_BLOCKED:
         on_send_blocked(faddr, prev_ps.path.remote, prev_ps.path.local, prev_pi,
-                        data, datalen, max_udp_payload_size);
+                        data, datalen, gso_size);
 
         on_send_blocked(static_cast<UpstreamAddr *>(ps.path.user_data),
                         ps.path.remote, ps.path.local, pi, bufpos - nwrite,
-                        nwrite, max_udp_payload_size);
+                        nwrite, 0);
 
         signal_write_upstream_addr(faddr);
 
@@ -899,40 +926,37 @@ int Http3Upstream::write_streams() {
 
         rv = send_packet(faddr, ps.path.remote.addr, ps.path.remote.addrlen,
                          ps.path.local.addr, ps.path.local.addrlen, pi, data,
-                         nwrite, max_udp_payload_size);
+                         nwrite, 0);
         if (rv == SHRPX_ERR_SEND_BLOCKED) {
           on_send_blocked(faddr, ps.path.remote, ps.path.local, pi, data,
-                          nwrite, max_udp_payload_size);
-        }
+                          nwrite, 0);
 
-        signal_write_upstream_addr(faddr);
+          signal_write_upstream_addr(faddr);
+        }
       }
       }
 
       ngtcp2_conn_update_pkt_tx_time(conn_, ts);
-      reset_idle_timer();
 
       return 0;
     }
 
-    if (++pktcnt == max_pktcnt ||
-        static_cast<size_t>(nwrite) < max_udp_payload_size) {
+    if (++pktcnt == max_pktcnt || static_cast<size_t>(nwrite) < gso_size) {
       auto faddr = static_cast<UpstreamAddr *>(ps.path.user_data);
       auto data = tx_.data.get();
       auto datalen = bufpos - data;
 
       rv = send_packet(faddr, ps.path.remote.addr, ps.path.remote.addrlen,
                        ps.path.local.addr, ps.path.local.addrlen, pi, data,
-                       datalen, max_udp_payload_size);
+                       datalen, gso_size);
       if (rv == SHRPX_ERR_SEND_BLOCKED) {
         on_send_blocked(faddr, ps.path.remote, ps.path.local, pi, data, datalen,
-                        max_udp_payload_size);
+                        gso_size);
+
+        signal_write_upstream_addr(faddr);
       }
 
       ngtcp2_conn_update_pkt_tx_time(conn_, ts);
-      reset_idle_timer();
-
-      signal_write_upstream_addr(faddr);
 
       return 0;
     }
@@ -949,7 +973,6 @@ int Http3Upstream::write_streams() {
                       0);
 
       ngtcp2_conn_update_pkt_tx_time(conn_, ts);
-      reset_idle_timer();
 
       signal_write_upstream_addr(faddr);
 
@@ -958,9 +981,6 @@ int Http3Upstream::write_streams() {
 
     if (++pktcnt == max_pktcnt) {
       ngtcp2_conn_update_pkt_tx_time(conn_, ts);
-      reset_idle_timer();
-
-      signal_write_upstream_addr(faddr);
 
       return 0;
     }
@@ -991,15 +1011,8 @@ int Http3Upstream::on_downstream_abort_request(Downstream *downstream,
 
 int Http3Upstream::on_downstream_abort_request_with_https_redirect(
     Downstream *downstream) {
-  int rv;
-
-  rv = redirect_to_https(downstream);
-  if (rv != 0) {
-    return -1;
-  }
-
-  handler_->signal_write();
-  return 0;
+  assert(0);
+  abort();
 }
 
 namespace {
@@ -1493,7 +1506,7 @@ void Http3Upstream::on_handler_delete() {
     quic_conn_handler->remove_connection_id(cid);
   }
 
-  if (idle_close_ || retry_close_) {
+  if (retry_close_ || last_error_.type == NGTCP2_CCERR_TYPE_IDLE_CLOSE) {
     return;
   }
 
@@ -1506,9 +1519,17 @@ void Http3Upstream::on_handler_delete() {
 
     ngtcp2_path_storage_zero(&ps);
 
+    ngtcp2_ccerr ccerr;
+    ngtcp2_ccerr_default(&ccerr);
+
+    if (worker->get_graceful_shutdown() &&
+        !ngtcp2_conn_get_handshake_completed(conn_)) {
+      ccerr.error_code = NGTCP2_CONNECTION_REFUSED;
+    }
+
     auto nwrite = ngtcp2_conn_write_connection_close(
-        conn_, &ps.path, &pi, conn_close_.data(), conn_close_.size(),
-        NGTCP2_NO_ERROR, nullptr, 0, quic_timestamp());
+        conn_, &ps.path, &pi, conn_close_.data(), conn_close_.size(), &ccerr,
+        quic_timestamp());
     if (nwrite < 0) {
       if (nwrite != NGTCP2_ERR_INVALID_STATE) {
         ULOG(ERROR, this) << "ngtcp2_conn_write_connection_close: "
@@ -1606,10 +1627,11 @@ int Http3Upstream::on_downstream_reset(Downstream *downstream, bool no_retry) {
 
 fail:
   if (rv == SHRPX_ERR_TLS_REQUIRED) {
-    rv = on_downstream_abort_request_with_https_redirect(downstream);
-  } else {
-    rv = on_downstream_abort_request(downstream, 502);
+    assert(0);
+    abort();
   }
+
+  rv = on_downstream_abort_request(downstream, 502);
   if (rv != 0) {
     shutdown_stream(downstream, NGHTTP3_H3_INTERNAL_ERROR);
   }
@@ -1766,51 +1788,40 @@ int Http3Upstream::on_read(const UpstreamAddr *faddr,
       auto worker = handler_->get_worker();
       auto quic_conn_handler = worker->get_quic_connection_handler();
 
-      uint32_t version;
-      const uint8_t *dcid, *scid;
-      size_t dcidlen, scidlen;
+      if (worker->get_graceful_shutdown()) {
+        ngtcp2_ccerr_set_transport_error(&last_error_,
+                                         NGTCP2_CONNECTION_REFUSED, nullptr, 0);
 
-      rv = ngtcp2_pkt_decode_version_cid(&version, &dcid, &dcidlen, &scid,
-                                         &scidlen, data, datalen,
-                                         SHRPX_QUIC_SCIDLEN);
-      if (rv != 0) {
-        return -1;
+        return handle_error();
       }
 
-      if (worker->get_graceful_shutdown()) {
-        ngtcp2_cid ini_dcid, ini_scid;
-
-        ngtcp2_cid_init(&ini_dcid, dcid, dcidlen);
-        ngtcp2_cid_init(&ini_scid, scid, scidlen);
-
-        quic_conn_handler->send_connection_close(
-            faddr, version, ini_dcid, ini_scid, remote_addr, local_addr,
-            NGTCP2_CONNECTION_REFUSED);
+      ngtcp2_version_cid vc;
 
+      rv =
+          ngtcp2_pkt_decode_version_cid(&vc, data, datalen, SHRPX_QUIC_SCIDLEN);
+      if (rv != 0) {
         return -1;
       }
 
       retry_close_ = true;
 
-      quic_conn_handler->send_retry(handler_->get_upstream_addr(), version,
-                                    dcid, dcidlen, scid, scidlen, remote_addr,
-                                    local_addr);
+      quic_conn_handler->send_retry(handler_->get_upstream_addr(), vc.version,
+                                    vc.dcid, vc.dcidlen, vc.scid, vc.scidlen,
+                                    remote_addr, local_addr, datalen * 3);
 
       return -1;
     }
-    case NGTCP2_ERR_REQUIRED_TRANSPORT_PARAM:
-    case NGTCP2_ERR_MALFORMED_TRANSPORT_PARAM:
-    case NGTCP2_ERR_TRANSPORT_PARAM:
-      // If rv indicates transport_parameters related error, we should
-      // send TRANSPORT_PARAMETER_ERROR even if last_error_.code is
-      // already set.  This is because OpenSSL might set Alert.
-      last_error_ = quic::err_transport(rv);
+    case NGTCP2_ERR_CRYPTO:
+      if (!last_error_.error_code) {
+        ngtcp2_ccerr_set_tls_alert(
+            &last_error_, ngtcp2_conn_get_tls_alert(conn_), nullptr, 0);
+      }
       break;
     case NGTCP2_ERR_DROP_CONN:
       return -1;
     default:
-      if (!last_error_.code) {
-        last_error_ = quic::err_transport(rv);
+      if (!last_error_.error_code) {
+        ngtcp2_ccerr_set_liberr(&last_error_, rv, nullptr, 0);
       }
     }
 
@@ -1819,8 +1830,6 @@ int Http3Upstream::on_read(const UpstreamAddr *faddr,
     return handle_error();
   }
 
-  reset_idle_timer();
-
   return 0;
 }
 
@@ -1838,7 +1847,7 @@ int Http3Upstream::send_packet(const UpstreamAddr *faddr,
     // large.
   case -EINVAL:
   case -EMSGSIZE:
-    max_udp_payload_size_ = NGTCP2_MAX_UDP_PAYLOAD_SIZE;
+    // Let the packet lost.
     break;
   case -EAGAIN:
 #if EAGAIN != EWOULDBLOCK
@@ -1857,7 +1866,7 @@ void Http3Upstream::on_send_blocked(const UpstreamAddr *faddr,
                                     const ngtcp2_addr &local_addr,
                                     const ngtcp2_pkt_info &pi,
                                     const uint8_t *data, size_t datalen,
-                                    size_t max_udp_payload_size) {
+                                    size_t gso_size) {
   assert(tx_.num_blocked || !tx_.send_blocked);
   assert(tx_.num_blocked < 2);
 
@@ -1874,7 +1883,7 @@ void Http3Upstream::on_send_blocked(const UpstreamAddr *faddr,
   p.pi = pi;
   p.data = data;
   p.datalen = datalen;
-  p.max_udp_payload_size = max_udp_payload_size;
+  p.gso_size = gso_size;
 }
 
 int Http3Upstream::send_blocked_packet() {
@@ -1887,7 +1896,7 @@ int Http3Upstream::send_blocked_packet() {
 
     rv = send_packet(p.faddr, &p.remote_addr.su.sa, p.remote_addr.len,
                      &p.local_addr.su.sa, p.local_addr.len, p.pi, p.data,
-                     p.datalen, p.max_udp_payload_size);
+                     p.datalen, p.gso_size);
     if (rv == SHRPX_ERR_SEND_BLOCKED) {
       signal_write_upstream_addr(p.faddr);
 
@@ -1917,7 +1926,8 @@ void Http3Upstream::signal_write_upstream_addr(const UpstreamAddr *faddr) {
 }
 
 int Http3Upstream::handle_error() {
-  if (ngtcp2_conn_is_in_closing_period(conn_)) {
+  if (ngtcp2_conn_is_in_closing_period(conn_) ||
+      ngtcp2_conn_is_in_draining_period(conn_)) {
     return -1;
   }
 
@@ -1930,30 +1940,21 @@ int Http3Upstream::handle_error() {
 
   conn_close_.resize(SHRPX_QUIC_CONN_CLOSE_PKTLEN);
 
-  ngtcp2_ssize nwrite;
-
-  if (last_error_.type == quic::ErrorType::Transport) {
-    nwrite = ngtcp2_conn_write_connection_close(
-        conn_, &ps.path, &pi, conn_close_.data(), conn_close_.size(),
-        last_error_.code, nullptr, 0, ts);
-    if (nwrite < 0) {
-      ULOG(ERROR, this) << "ngtcp2_conn_write_connection_close: "
-                        << ngtcp2_strerror(nwrite);
-      return -1;
-    }
-  } else {
-    nwrite = ngtcp2_conn_write_application_close(
-        conn_, &ps.path, &pi, conn_close_.data(), conn_close_.size(),
-        last_error_.code, nullptr, 0, ts);
-    if (nwrite < 0) {
-      ULOG(ERROR, this) << "ngtcp2_conn_write_application_close: "
-                        << ngtcp2_strerror(nwrite);
-      return -1;
-    }
+  auto nwrite = ngtcp2_conn_write_connection_close(
+      conn_, &ps.path, &pi, conn_close_.data(), conn_close_.size(),
+      &last_error_, ts);
+  if (nwrite < 0) {
+    ULOG(ERROR, this) << "ngtcp2_conn_write_connection_close: "
+                      << ngtcp2_strerror(nwrite);
+    return -1;
   }
 
   conn_close_.resize(nwrite);
 
+  if (nwrite == 0) {
+    return -1;
+  }
+
   send_packet(static_cast<UpstreamAddr *>(ps.path.user_data),
               ps.path.remote.addr, ps.path.remote.addrlen, ps.path.local.addr,
               ps.path.local.addrlen, pi, conn_close_.data(), nwrite, 0);
@@ -1961,47 +1962,6 @@ int Http3Upstream::handle_error() {
   return -1;
 }
 
-int Http3Upstream::on_rx_secret(ngtcp2_crypto_level level,
-                                const uint8_t *secret, size_t secretlen) {
-  if (ngtcp2_crypto_derive_and_install_rx_key(conn_, nullptr, nullptr, nullptr,
-                                              level, secret, secretlen) != 0) {
-    ULOG(ERROR, this) << "ngtcp2_crypto_derive_and_install_rx_key failed";
-    return -1;
-  }
-
-  return 0;
-}
-
-int Http3Upstream::on_tx_secret(ngtcp2_crypto_level level,
-                                const uint8_t *secret, size_t secretlen) {
-  if (ngtcp2_crypto_derive_and_install_tx_key(conn_, nullptr, nullptr, nullptr,
-                                              level, secret, secretlen) != 0) {
-    ULOG(ERROR, this) << "ngtcp2_crypto_derive_and_install_tx_key failed";
-    return -1;
-  }
-
-  if (level == NGTCP2_CRYPTO_LEVEL_APPLICATION && setup_httpconn() != 0) {
-    return -1;
-  }
-
-  return 0;
-}
-
-int Http3Upstream::add_crypto_data(ngtcp2_crypto_level level,
-                                   const uint8_t *data, size_t datalen) {
-  int rv = ngtcp2_conn_submit_crypto_data(conn_, level, data, datalen);
-
-  if (rv != 0) {
-    ULOG(ERROR, this) << "ngtcp2_conn_submit_crypto_data: "
-                      << ngtcp2_strerror(rv);
-    return -1;
-  }
-
-  return 0;
-}
-
-void Http3Upstream::set_tls_alert(uint8_t alert) { tls_alert_ = alert; }
-
 int Http3Upstream::handle_expiry() {
   int rv;
 
@@ -2009,34 +1969,31 @@ int Http3Upstream::handle_expiry() {
 
   rv = ngtcp2_conn_handle_expiry(conn_, ts);
   if (rv != 0) {
-    ULOG(ERROR, this) << "ngtcp2_conn_handle_expiry: " << ngtcp2_strerror(rv);
-    last_error_ = quic::err_transport(rv);
+    if (rv == NGTCP2_ERR_IDLE_CLOSE) {
+      ULOG(INFO, this) << "Idle connection timeout";
+    } else {
+      ULOG(ERROR, this) << "ngtcp2_conn_handle_expiry: " << ngtcp2_strerror(rv);
+    }
+    ngtcp2_ccerr_set_liberr(&last_error_, rv, nullptr, 0);
     return handle_error();
   }
 
   return 0;
 }
 
-void Http3Upstream::reset_idle_timer() {
-  auto ts = quic_timestamp();
-  auto idle_ts = ngtcp2_conn_get_idle_expiry(conn_);
-
-  idle_timer_.repeat =
-      idle_ts > ts ? static_cast<ev_tstamp>(idle_ts - ts) / NGTCP2_SECONDS
-                   : 1e-9;
-
-  ev_timer_again(handler_->get_loop(), &idle_timer_);
-}
-
 void Http3Upstream::reset_timer() {
   auto ts = quic_timestamp();
   auto expiry_ts = ngtcp2_conn_get_expiry(conn_);
+  auto loop = handler_->get_loop();
 
-  timer_.repeat = expiry_ts > ts
-                      ? static_cast<ev_tstamp>(expiry_ts - ts) / NGTCP2_SECONDS
-                      : 1e-9;
+  if (expiry_ts <= ts) {
+    ev_feed_event(loop, &timer_, EV_TIMER);
+    return;
+  }
 
-  ev_timer_again(handler_->get_loop(), &timer_);
+  timer_.repeat = static_cast<ev_tstamp>(expiry_ts - ts) / NGTCP2_SECONDS;
+
+  ev_timer_again(loop, &timer_);
 }
 
 namespace {
@@ -2251,11 +2208,11 @@ int Http3Upstream::http_end_request_headers(Downstream *downstream, int fin) {
 
   auto content_length = req.fs.header(http2::HD_CONTENT_LENGTH);
   if (content_length) {
-    // libnghttp2 guarantees this can be parsed
+    // libnghttp3 guarantees this can be parsed
     req.fs.content_length = util::parse_uint(content_length->value);
   }
 
-  // presence of mandatory header fields are guaranteed by libnghttp2.
+  // presence of mandatory header fields are guaranteed by libnghttp3.
   auto authority = req.fs.header(http2::HD__AUTHORITY);
   auto path = req.fs.header(http2::HD__PATH);
   auto method = req.fs.header(http2::HD__METHOD);
@@ -2369,16 +2326,19 @@ void Http3Upstream::start_downstream(Downstream *downstream) {
 void Http3Upstream::initiate_downstream(Downstream *downstream) {
   int rv;
 
+#ifdef HAVE_MRUBY
   DownstreamConnection *dconn_ptr;
+#endif // HAVE_MRUBY
 
   for (;;) {
     auto dconn = handler_->get_downstream_connection(rv, downstream);
     if (!dconn) {
       if (rv == SHRPX_ERR_TLS_REQUIRED) {
-        rv = redirect_to_https(downstream);
-      } else {
-        rv = error_reply(downstream, 502);
+        assert(0);
+        abort();
       }
+
+      rv = error_reply(downstream, 502);
       if (rv != 0) {
         shutdown_stream(downstream, NGHTTP3_H3_INTERNAL_ERROR);
       }
@@ -2625,7 +2585,7 @@ int Http3Upstream::http_reset_stream(int64_t stream_id,
 int Http3Upstream::setup_httpconn() {
   int rv;
 
-  if (ngtcp2_conn_get_max_local_streams_uni(conn_) < 3) {
+  if (ngtcp2_conn_get_streams_uni_left(conn_) < 3) {
     return -1;
   }
 
@@ -2663,11 +2623,10 @@ int Http3Upstream::setup_httpconn() {
     return -1;
   }
 
-  ngtcp2_transport_params params;
-  ngtcp2_conn_get_local_transport_params(conn_, &params);
+  auto params = ngtcp2_conn_get_local_transport_params(conn_);
 
   nghttp3_conn_set_max_client_streams_bidi(httpconn_,
-                                           params.initial_max_streams_bidi);
+                                           params->initial_max_streams_bidi);
 
   int64_t ctrl_stream_id;
 
@@ -2789,39 +2748,6 @@ int Http3Upstream::shutdown_stream_read(int64_t stream_id,
   return 0;
 }
 
-int Http3Upstream::redirect_to_https(Downstream *downstream) {
-  auto &req = downstream->request();
-  if (req.regular_connect_method() || req.scheme != "http") {
-    return error_reply(downstream, 400);
-  }
-
-  auto authority = util::extract_host(req.authority);
-  if (authority.empty()) {
-    return error_reply(downstream, 400);
-  }
-
-  auto &balloc = downstream->get_block_allocator();
-  auto config = get_config();
-  auto &httpconf = config->http;
-
-  StringRef loc;
-  if (httpconf.redirect_https_port == StringRef::from_lit("443")) {
-    loc = concat_string_ref(balloc, StringRef::from_lit("https://"), authority,
-                            req.path);
-  } else {
-    loc = concat_string_ref(balloc, StringRef::from_lit("https://"), authority,
-                            StringRef::from_lit(":"),
-                            httpconf.redirect_https_port, req.path);
-  }
-
-  auto &resp = downstream->response();
-  resp.http_status = 308;
-  resp.fs.add_header_token(StringRef::from_lit("location"), loc, false,
-                           http2::HD_LOCATION);
-
-  return send_reply(downstream, nullptr, 0);
-}
-
 void Http3Upstream::consume(int64_t stream_id, size_t nconsumed) {
   ngtcp2_conn_extend_max_stream_offset(conn_, stream_id, nconsumed);
   ngtcp2_conn_extend_max_offset(conn_, nconsumed);
@@ -2878,6 +2804,10 @@ int Http3Upstream::start_graceful_shutdown() {
     return 0;
   }
 
+  if (!httpconn_) {
+    return -1;
+  }
+
   rv = nghttp3_conn_submit_shutdown_notice(httpconn_);
   if (rv != 0) {
     ULOG(FATAL, this) << "nghttp3_conn_submit_shutdown_notice: "
@@ -2910,8 +2840,6 @@ int Http3Upstream::submit_goaway() {
   return 0;
 }
 
-void Http3Upstream::idle_close() { idle_close_ = true; }
-
 int Http3Upstream::open_qlog_file(const StringRef &dir,
                                   const ngtcp2_cid &scid) const {
   std::array<char, sizeof("20141115T125824.741+0900")> buf;
@@ -2952,4 +2880,6 @@ int Http3Upstream::open_qlog_file(const StringRef &dir,
   return fd;
 }
 
+ngtcp2_conn *Http3Upstream::get_conn() const { return conn_; }
+
 } // namespace shrpx
